%{
/* shapeparser.y
 *
 * Yacc script for parsing shapes into the sim 3d representation. Used only
 * in the server
 */
#include "sim.h"
#include "server.h"

static void write_shape_base(int fd, int tl, shape *sh, int lvl, int toserver);
static char *pointstr(point p);
%}
%union {
	double num;
	char string[256];
	shapecol col;
	}
%token <num> NUM
%token <string> STRING
%token END,POLY,SPHERE,BOX,CONE,TUBE,COLL,POS,NAME,COLOUR,DISC,COPY,HIDDEN
%token ROTATE,CENTRE,RADIUS,HEIGHT,POINTS,CORNERS,SURFACE,SCALE,TEXTURE,CMT
%type <col> colour
%type <string> opt_name strings
%type <num> surftype
%%
input		:	END
			{ return -1;
			}
		|	shape
			{ return 0;
			}
		;

shape		:	poly
		|	sphere
		|	box
		|	cone
		|	tube
		|	disc
		|	copy
		|	coll
		;

shape_in_coll	:	{ /* Add current shape (collection) to stack
			   */
			  coll[collp++] = ns;
			}
			shape
			{ /* Add to collection list
			   */
			  ns->next = coll[collp-1]->shp.coll.head;
			  coll[collp-1]->shp.coll.head = ns;

			  /* Set current shape to collection from stack
			   */
			  ns = coll[--collp];
			}
		;


/* poly - a shape consisting of one or more planes
 */
poly		:	POLY opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_poly;
			}
			'{' polyattribs '}'
		;

opt_name	:	STRING { strcpy($$,$1); }
		|	{ $$[0] = 0; }
		;

polyattribs	:	polyattribs polyattrib
		|	polyattribs shapeattrib
		|
		;

polyattrib	:	POINTS { ptc_temp = 0; } polypoints
			{ if (ptc_temp < 3) {
			    yyerror("Polygon must have >= 3 points");
			    return 1;
			    }
			  if (ns->shp.poly.facec >= MAX_FACES) {
			    yyerror("Polygon has too many faces");
			    return 1;
			    }
			  ns->shp.poly.pt[ns->shp.poly.facec] =
			   memdupe(pt_temp,ptc_temp*sizeof(point));
			  ns->shp.poly.ptc[ns->shp.poly.facec] = ptc_temp;
			  ns->shp.poly.facec++;
			}
		;

polypoints	:	polypoints point
		|
		;

point		:	NUM NUM NUM
			{ pt_temp[ptc_temp].x = $1;
			  pt_temp[ptc_temp].y = $2;
			  pt_temp[ptc_temp].z = $3;
			  ptc_temp++;
			}
		;


/* sphere - a single sphere 
 */
sphere		:	SPHERE opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_sphere;
			  ns->shp.sphere.rad   = 0.5;
			  ns->shp.sphere.cen.x = 0.0;
			  ns->shp.sphere.cen.y = 0.0;
			  ns->shp.sphere.cen.z = 0.0;
			}
			'{' sphereattribs '}'
		;

sphereattribs	:	sphereattribs sphereattrib
		|	sphereattribs shapeattrib
		|
		;

sphereattrib	:	CENTRE NUM NUM NUM
			{ ns->shp.sphere.cen.x = $2;
			  ns->shp.sphere.cen.y = $3;
			  ns->shp.sphere.cen.z = $4;
			}
		|	RADIUS NUM
			{ ns->shp.sphere.rad = $2;
			}
		;


/* box - a cuboid initially aligned along the exis
 */
box		:	BOX opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_box;
			  ns->shp.box.c1.x = -0.5;
			  ns->shp.box.c1.y = -0.5;
			  ns->shp.box.c1.z = -0.5;
			  ns->shp.box.c2.x = 0.5;
			  ns->shp.box.c2.y = 0.5;
			  ns->shp.box.c2.z = 0.5;
			}
			'{' boxattribs '}'
		;

boxattribs	:	boxattribs boxattrib
		|	boxattribs shapeattrib
		|
		;

boxattrib	:	CORNERS  NUM NUM NUM  NUM NUM NUM
			{ ns->shp.box.c1.x = $2;
			  ns->shp.box.c1.y = $3;
			  ns->shp.box.c1.z = $4;
			  ns->shp.box.c2.x = $5;
			  ns->shp.box.c2.y = $6;
			  ns->shp.box.c2.z = $7;
			}
		;


/* cone - a single cone initially standing vertically
 */
cone		:	CONE opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_cone;
			  ns->shp.cone.rad    = 0.5;
			  ns->shp.cone.base.x = 0.0;
			  ns->shp.cone.base.y = 0.0;
			  ns->shp.cone.base.z = 0.0;
			  ns->shp.cone.height = 1.0;
			}
			'{' coneattribs '}'
		;

coneattribs	:	coneattribs coneattrib
		|	coneattribs shapeattrib
		|
		;

coneattrib	:	CENTRE NUM NUM NUM
			{ ns->shp.cone.base.x = $2;
			  ns->shp.cone.base.y = $3;
			  ns->shp.cone.base.z = $4;
			}
		|	RADIUS NUM
			{ ns->shp.cone.rad = $2;
			}
		|	HEIGHT NUM
			{ ns->shp.cone.height = $2;
			}
		;


/* tube - a single tube initially standing vertically
 */
tube		:	TUBE opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_tube;
			  ns->shp.tube.rad    = 0.5;
			  ns->shp.tube.base.x = 0.0;
			  ns->shp.tube.base.y = 0.0;
			  ns->shp.tube.base.z = 0.0;
			  ns->shp.tube.height = 1.0;
			}
			'{' tubeattribs '}'
		;

tubeattribs	:	tubeattribs tubeattrib
		|	tubeattribs shapeattrib
		|
		;

tubeattrib	:	CENTRE NUM NUM NUM
			{ ns->shp.tube.base.x = $2;
			  ns->shp.tube.base.y = $3;
			  ns->shp.tube.base.z = $4;
			}
		|	RADIUS NUM
			{ ns->shp.tube.rad = $2;
			}
		|	HEIGHT NUM
			{ ns->shp.tube.height = $2;
			}
		;


/* disc - a single disc, initally aligned in the xz plane
 */
disc		:	DISC opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_disc;
			  ns->shp.disc.cen.x = 0.0;
			  ns->shp.disc.cen.y = 0.0;
			  ns->shp.disc.cen.z = 0.0;
			  ns->shp.disc.rad   = 0.5;
			}
			'{' discattribs '}'
		;

discattribs	:	discattribs discattrib
		|	discattribs shapeattrib
		|
		;

discattrib	:	CENTRE NUM NUM NUM
			{ ns->shp.disc.cen.x = $2;
			  ns->shp.disc.cen.y = $3;
			  ns->shp.disc.cen.z = $4;
			}
		|	RADIUS NUM
			{ ns->shp.disc.rad = $2;
			}
		;


/* coll - a collection of other shapes
 */
coll		:	COLL opt_name
			{ ns = create_shape();
			  strcpy(ns->name,$2);
			  ns->type = st_coll;
			  /* Shapes inside this one are added to its own
			   * internal list only
			   */
			  ns->shp.coll.head = NULL;
			}
			'{' collattribs '}'
		;

collattribs	:	collattribs shape_in_coll
		|	collattribs shapeattrib
		|
		;


/* copy - a deep copy of any other previously defined shape
 */
copy		:	COPY STRING opt_name
			{ shape *src;

			  if ((src = lookup_shape($2)) == NULL) {
			    yyerror("Source shape not found");
			    return 1;
			    }
			  ns = copy_shape(src);

			  strcpy(ns->name,$3);
			}
			'{' copyattribs '}'
		;

copyattribs	:	copyattribs shapeattrib
		|
		;

/* When specifying the colour of a shape, either the name can be used
 * or a red-green-blue triple
 */
colour		:	NUM NUM NUM
			{ shapecol col;
			  col.red = $1;
			  col.green = $2;
			  col.blue = $3;
			  if (col.red < 0 || col.red > 255 ||
			      col.green < 0 || col.green > 255 ||
			      col.blue < 0 || col.blue > 255) {
			    yyerror("RGB values must be between 0 and 255");
			    return 1;
			    }
			  $$ = col;
			}
		;

shapeattrib	:	NAME STRING
			{ strcpy(ns->name,$2);
			}
		|	POS NUM NUM NUM
			{ ns->pos.x = $2;
			  ns->pos.y = $3;
			  ns->pos.z = $4;
			}
		|	COLOUR colour
			{ ns->colour = $2;
			}
		|	HIDDEN NUM
			{ ns->hidden = $2;
			  /* for stuff in inventory not to be sent to clients
			   */
			}
		|	ROTATE NUM  NUM NUM NUM
			{ ns->rota = $2;
			  ns->rotx = $3;
			  ns->roty = $4;
			  ns->rotz = $5;
			}
		|	SURFACE surftype
			{ ns->surf = $2;
			}
		|	SCALE NUM NUM NUM
			{ ns->scx = $2;
			  ns->scy = $3;
			  ns->scz = $4;
			}
		|	TEXTURE STRING
			{ char fname[256];
			  int ppmfd;

			  if (ns->tex.data)
			    free(ns->tex.data);
			  if ($2[0] == '/')	/* absolute path */
			    sprintf(fname,"%s",$2);
			  else			/* relative to map dir */
			    sprintf(fname,"%s/%s",mp->mapdir,$2);
			  if ((ppmfd = open(fname,O_RDONLY)) == -1) {
			    yyerror("ppm file not found");
			    return 1;
			    }
			  ns->tex.data = read_ppm_texture(ppmfd,&ns->tex.w,
							  &ns->tex.h);
			  close(ppmfd);
			  if (!ns->tex.data) {
			    yyerror("Error reading texture file");
			    return 1;
			    }
			}
		|	TEXTURE NUM NUM
			{ if (ns->tex.data)
			    free(ns->tex.data);
			  ns->tex.w = $2;
			  ns->tex.h = $3;
			  ns->tex.data = scalloc(1,ns->tex.w * ns->tex.h * 3);
			  texc = 0;
			}
			'{' texture_strings '}'
			{ if (texc != ns->tex.w * ns->tex.h * 3) {
			    yyerror("Incorrect number of texture values");
			    return 1;
			    }
			}
		|	CMT strings
			{ if (ns->cmt)
			    free(ns->cmt);
			  ns->cmt = strdupe($2);
			} 
		;

texture_strings	:	texture_strings texture_string
		|
		;

texture_string	:	STRING
			{ int len = strlen($1);
			  decode_rgb($1,&ns->tex.data[texc],len);
			  texc += (len/4)*3;
			}

strings		:	strings STRING
			{ sprintf($$,"%s%s",$1,$2);
			}
		|	STRING
			{ sprintf($$,"%s",$1);
			}
		;

surftype	:	NUM
			{ $$ = $1; }
		|	STRING
			{ if (strequ($1,"flat"))
			    $$ = sft_flat;
			  else if (strequ($1,"glass"))
			    $$ = sft_glass;
			  else if (strequ($1,"shiny"))
			    $$ = sft_shiny;
			  else {
			    yyerror("Unknown surface type");
			    return 1;
			    }
			}
		;

%%
static shape *ns;		/* shape currently being created */
static shape *coll[256];	/* stack of collections */
static int    collp;		/* number of collections on stack */
static point pt_temp[256];
static int ptc_temp;
static int texc;

static reserved rwtab[] = {	"end",END,"poly",POLY,"sphere",SPHERE,
				"box",BOX,"cone",CONE,"copy",COPY,
				"tube",TUBE,"disc",DISC,"coll",COLL,
				"pos",POS,"name",NAME,"colour",COLOUR,
				"hidden",HIDDEN,"rotate",ROTATE,
				"centre",CENTRE,"radius",RADIUS,
				"height",HEIGHT,"points",POINTS,
				"corners",CORNERS,"surface",SURFACE,
				"scale",SCALE,"texture",TEXTURE,
				"cmt",CMT };

#define error yyerror
#define lex yylex
#define lval yylval

#include "lexer.h"

/* parse_shape
 * Read one shape from a file descriptor. Returns 0 if ok, 1 on error and
 * -1 if 'end' was read.
 */
int parse_shape(int filed, shape **sptr)
{
int rv;

init_lexer(filed);
collp = 0;
rv = yyparse();
*sptr = ns;
return rv;
}


/* parse_shapes
 * Read shapes from a file descriptor and add them to the shape table
 */
int parse_shapes(int filed)
{
int rv;

init_lexer(filed);
collp = 0;
while((rv = yyparse()) == 0)
	add_shape(ns);
if (rv == 1)
	return 1;
else
	return 0;
}


/* write_shape
 * Output one shape to a file descriptor.
 */
void write_shape(int fd, int tl, shape *sh, int lvl, int toserver)
{
shape *sp;
int i;

/* Output 'base' information
 */
write_shape_base(fd,tl,sh,lvl,toserver);

/* Output other attributes
 */
if (sh->pos.x != 0.0 || sh->pos.y != 0.0 || sh->pos.z != 0.0)
	fdprintf(fd,"%spos %s\n",tabs(tl+1),pointstr(sh->pos));
if (!(sh->type == st_coll ||
     (sh->type == st_copy && sh->shp.copy.src->type == st_coll)))
	fdprintf(fd,"%scolour %d %d %d\n",tabs(tl+1),sh->colour.red,
						     sh->colour.green,
				  		     sh->colour.blue);
if (sh->name[0])
	fdprintf(fd,"%sname \"%s\"\n",tabs(tl+1),sh->name);
if (sh->hidden)
	fdprintf(fd,"%shidden %d\n",tabs(tl+1),sh->hidden);
if (sh->rota)
	fdprintf(fd,"%srotate %f %f %f %f\n",tabs(tl+1),sh->rota,sh->rotx,
					     sh->roty,sh->rotz);
if (sh->surf != sft_flat)
	fdprintf(fd,"%ssurface %d\n",tabs(tl+1),sh->surf);
if (sh->scx != 1.0 || sh->scy != 1.0 || sh->scz != 1.0)
	fdprintf(fd,"%sscale %f %f %f\n",tabs(tl+1),sh->scx,sh->scy,sh->scz);
if (toserver != 0 && sh->cmt)
	fdprintf(fd,"%scmt \"%s\"\n",tabs(tl+1),sh->cmt);
if (sh->tex.data && lvl == lvl_textured) {
	fdprintf(fd,"%stexture %d %d {\n",tabs(tl+1),sh->tex.w,sh->tex.h);
	write_encoded_texture(fd,sh->tex.data,sh->tex.w,sh->tex.h);
	fdprintf(fd,"%s}\n",tabs(tl+2));
	}
fdprintf(fd,"%s}\n",tabs(tl+1));
}


/* write_shape_base
 * Output the base geometric information about a shape
 */
static void write_shape_base(int fd, int tl, shape *sh, int lvl, int toserver)
{
int i;
shape *sp;

switch(sh->type) {
case st_poly:
	fdprintf(fd,"%spoly {\n",tabs(tl));
	for(i=0; i<sh->shp.poly.facec; i++) {
		int j;

		fdprintf(fd,"%spoints",tabs(tl+1));
		for(j=0; j<sh->shp.poly.ptc[i]; j++)
			fdprintf(fd," %s ",pointstr(sh->shp.poly.pt[i][j]));
		fdprintf(fd,"\n");
		}
	break;
case st_sphere:
	fdprintf(fd,"%ssphere {\n",tabs(tl));
	fdprintf(fd,"%sradius %f\n",tabs(tl+1),sh->shp.sphere.rad);
	fdprintf(fd,"%scentre %s\n",tabs(tl+1),pointstr(sh->shp.sphere.cen));
	break;
case st_box:
	fdprintf(fd,"%sbox {\n",tabs(tl));
	fdprintf(fd,"%scorners %s  ",tabs(tl+1),pointstr(sh->shp.box.c1));
	fdprintf(fd,"%s\n",pointstr(sh->shp.box.c2));
	break;
case st_cone:
	fdprintf(fd,"%scone {\n",tabs(tl));
	fdprintf(fd,"%sradius %f\n",tabs(tl+1),sh->shp.cone.rad);
	fdprintf(fd,"%sheight %f\n",tabs(tl+1),sh->shp.cone.height);
	fdprintf(fd,"%scentre %s\n",tabs(tl+1),pointstr(sh->shp.cone.base));
	break;
case st_tube:
	fdprintf(fd,"%stube {\n",tabs(tl));
	fdprintf(fd,"%sradius %f\n",tabs(tl+1),sh->shp.tube.rad);
	fdprintf(fd,"%sheight %f\n",tabs(tl+1),sh->shp.tube.height);
	fdprintf(fd,"%scentre %s\n",tabs(tl+1),pointstr(sh->shp.tube.base));
	break;
case st_disc:
	fdprintf(fd,"%sdisc {\n",tabs(tl));
	fdprintf(fd,"%sradius %f\n",tabs(tl+1),sh->shp.disc.rad);
	fdprintf(fd,"%scentre %s\n",tabs(tl+1),pointstr(sh->shp.disc.cen));
	break;
case st_coll:
	fdprintf(fd,"%scoll {\n",tabs(tl));
	for(sp=sh->shp.coll.head; sp; sp=sp->next)
		write_shape(fd,tl+1,sp,lvl,toserver);
	break;
case st_copy:
	if (toserver == 2) {
		/* Just output a reference to the original shape, to save
		 * space in server dump files.
		 */
		fdprintf(fd,"%scopy \"%s\" {\n",
			    tabs(tl),sh->shp.copy.src->name);
		}
	else {
		/* Output the base details of the COPIED shape, which will
		 * then be followed by other attributes.
		 */
		write_shape_base(fd,tl,sh->shp.copy.src,lvl,toserver);
		}
	}
}


/* write_shapes
 * Output all shapes in the hash table, except ones starting with '_' unless
 * write_defs is non-zero.
 */
void write_shapes(int fd, int lvl, int toserver)
{
shape *s;

reset_shape();
while((s = next_shape()))
	if ((s->name[0] != '_' && !s->hidden) || toserver == 2) {
		write_shape(fd,0,s,lvl,toserver);
		fdputc(fd,'\n');
		}
fdprintf(fd,"end\n");
}


/* pointstr
 * Convert a point struct to a string
 */
static char *pointstr(point p)
{
static char buf[256];

sprintf(buf,"%f %f %f",p.x,p.y,p.z);
return buf;
}


/* write_all_shapes
 * Stub to output ALL shapes to a file descriptor
 */
void write_all_shapes(int fd)
{
write_shapes(fd,lvl_textured,2);
}


/* write_one_shape
 * Stub to output one shape to a file descriptor
 */
void write_one_shape(int fd, int tl, shape *sh)
{
write_shape(fd,tl,sh,lvl_textured,2);
}

